import numpy as np
import scipy.stats

from mixem.distribution.distribution import Distribution


class NormalDistribution(Distribution):
    """Univariate normal distribution with parameters (mu, sigma)."""

    @staticmethod
    def density(data, parameters):
        mu, sigma = parameters
        assert(len(data.shape) == 1), "Expect 1D data!"

        return scipy.stats.norm.pdf(data, mu, sigma)

    @staticmethod
    def estimate_parameters(data, weights):
        assert(len(data.shape) == 1), "Expect 1D data!"

        wsum = np.sum(weights)

        mu = np.sum(weights * data) / wsum
        sigma = np.sqrt(np.sum(weights * (data - mu) ** 2) / wsum)

        return mu, sigma


class MultivariateNormalDistribution(Distribution):
    """Multivariate normal distribution with parameters (mu, Sigma)."""

    @staticmethod
    def density(data, parameters):
        mu, sigma = parameters
        return scipy.stats.multivariate_normal.pdf(data, mu, sigma)

    @staticmethod
    def estimate_parameters(data, weights):
        mu = np.sum(data * weights[:, np.newaxis], axis=0) / np.sum(weights)

        center_x = data - mu[np.newaxis, :]

        # sigma = (np.diag(weights) @ center_x).T @ center_x / np.sum(weights)
        sigma = np.dot(
            np.dot(
                np.diag(weights),
                center_x
            ).T,
            center_x
        ) / np.sum(weights)

        return mu, sigma



import numpy as np

from mixem.distribution.distribution import Distribution


class ExponentialDistribution(Distribution):
    """Exponential distribution with parameter (lambda)."""

    @staticmethod
    def density(data, lmbda):
        assert(len(data.shape) == 1), "Expect 1D data!"
        return lmbda * np.exp(-lmbda * data)

    @staticmethod
    def estimate_parameters(data, weights):
        assert(len(data.shape) == 1), "Expect 1D data!"
        lmbda = np.sum(weights) / np.sum(weights * data)
        return lmbda



import abc
import numpy as np


class Distribution(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def density(self, data, parameters):
        return np.ones((data.shape[0],))

    @abc.abstractmethod
    def estimate_parameters(self, data, weights):
        return [], 0



import numpy as np


def em(data, distributions, initial_weights, initial_parameters, max_iterations=100, tol=1e-15, tol_iters=10, progress_callback=None):

    n_distr = len(distributions)
    n_data = data.shape[0]

    weight = np.array(initial_weights)
    param = initial_parameters

    last_ll = np.zeros((tol_iters, ))
    resp = np.empty((n_data, n_distr))
    density = np.empty((n_data, n_distr))

    iteration = 0
    while True:
        # E-step #######

        # compute responsibilities
        for d in range(n_distr):
            density[:, d] = distributions[d].density(data, param[d])

        # normalize responsibilities of distributions so they sum up to one for example
        resp = weight[np.newaxis, :] * density
        resp /= np.sum(resp, axis=1)[:, np.newaxis]

        log_likelihood = np.sum(resp * np.log(density))

        # M-step #######
        for d in range(n_distr):
            param[d] = distributions[d].estimate_parameters(data, resp[:, d])

        weight = np.mean(resp, axis=0)

        if progress_callback:
            progress_callback(iteration, weight, param, log_likelihood)

        # Convergence check #######
        if np.isnan(log_likelihood):
            last_ll[0] = log_likelihood
            break

        if iteration >= tol_iters and (last_ll[-1] - log_likelihood) / last_ll[-1] <= tol:
            last_ll[0] = log_likelihood
            break

        if iteration >= max_iterations:
            last_ll[0] = log_likelihood
            break

        # store value of current iteration in last_ll[0]
        # and shift older values to the right
        last_ll[1:] = last_ll[:-1]
        last_ll[0] = log_likelihood

        iteration += 1

    return weight, param, last_ll[0]



import numpy as np

from mixem.distribution.distribution import Distribution


class SimpleDistribution(Distribution):
    def __init__(self, f_density, f_estimate_parameters):
        self.f_density = f_density,
        self.f_estimate_parameters = f_estimate_parameters

    def density(self, *args, **kwargs):
        return self.f_density(*args, **kwargs)
 
    def estimate_parameters(self, *args, **kwargs):
        return self.f_estimate_parameters(*args, **kwargs)



import os
from setuptools import setup


def read(fname):
    try:
        with open(os.path.join(os.path.dirname(__file__), fname)) as f:
            return f.read()
    except:
        return ""

setup(
    name="mixem",
    version="0.0.4",

    author="Stefan Seemayer",
    author_email="stefan@seemayer.de",

    description="Expectation-Maximization (EM) algorithm for fitting mixtures of probability distributions",
    long_description=read("README.rst"),

    license="MIT",
    keywords="numeric em expectation maximization probability statistics distribution",
    url="https://github.com/sseemayer/mixem",

    packages=['mixem', 'mixem.distribution'],

    install_requires=[
        "numpy>=1.7.0",
        "scipy>=0.14.0",
    ],

    classifiers=[
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",

        "Intended Audience :: Science/Research",
        "Intended Audience :: Financial and Insurance Industry",

        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
)



from mixem.distribution.distribution import Distribution
from mixem.distribution.simple import SimpleDistribution
from mixem.distribution.normal import NormalDistribution, MultivariateNormalDistribution
from mixem.distribution.exponential import ExponentialDistribution
